import gzip
import io
import os
import hashlib
import json
import tarfile
import codecs
import shutil

from . import default
from . import util


class ArchiveReader(object):
    def __init__(self, path):
        self.path = path
        self.archive = None

        if os.path.isdir(path):
            self.tar = None
            self.meta = util.json_load(os.path.join(path, default.META_FILENAME))
            self.archive = io.open(os.path.join(path, self.filename()), 'rb')

        else:  # expect tar
            self.tar = tarfile.open(self.path, 'r')
            self.meta = self.get_meta()
            self.archive = self.tar.extractfile(self.filename())

    def filename(self):
        return os.path.basename(self.meta['archive'][0])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if isinstance(exc_value, Exception):
            return False
        self.close()

    def close(self):
        self.archive.close()
        self.tar.close()

    def extract(self, member, extract_path, cb=None):
        noffset = 0
        for entry in self.meta['manifest']:
            if entry['path'] == member:
                path = os.path.sep.join(entry['path'])
                size = entry['size']

                self.archive.seek(noffset)

                checksum = getattr(hashlib, entry['checksum'][0])()

                gz = gzip.GzipFile(fileobj=self.archive)

                filename = os.path.join(extract_path, path)
                util.makedirs(filename)
                with io.open(filename, 'wb') as f:

                    bytes_read = 0
                    while True:
                        chunk = gz.read(min(size - bytes_read, default.CHUNK_SIZE))
                        if not chunk:
                            break

                        bytes_read += len(chunk)

                        f.write(chunk)
                        checksum.update(chunk)

                        # callback for progress tracking
                        if cb:
                            cb(bytes_read)

                gz.close()

                # checksums from bytes read and meta data should match
                if checksum.hexdigest() != entry['checksum'][1]:
                    raise Exception('checksum mismatch: %s' % path)

            noffset = entry['noffset']

    def extract_all(self, extract_path, cb=None):
        for entry in self.meta['manifest']:
            self.extract(entry['path'], extract_path, cb=cb)

        members = [m['name'] for m in self.index_members()]
        for member in members:
            if self.tar:
                self.tar.extract(member, extract_path)
            else:
                shutil.copy(os.path.join(self.path, member), extract_path)

    def get_meta(self):
        reader = codecs.getreader('utf8')
        return json.load(reader(self.tar.extractfile(default.META_FILENAME)))

    def get_member(self, member):
        return self.meta[member]

    def list(self):
        return [tuple(e['path']) for e in self.meta['manifest']]

    def size_compressed(self):
        return self.meta[-1]['noffset']

    def index_members(self):
        if self.tar:
            return [{
                'size': m.size,
                'name': m.name,
            } for m in self.tar.getmembers() if m.name != default.ARCHIVE_FILENAME]
        else:
            return [{
                'size': os.stat(os.path.join(self.path, m)).st_size,
                'name': m,
            } for m in os.listdir(self.path) if m != default.ARCHIVE_FILENAME]

    def size(self):
        indices = [m['size'] for m in self.index_members()]
        files = [m['size'] for m in self.meta['manifest']]
        return sum(indices) + sum(files)



import os
import sys

from . import cli


def main():
    parser = cli.get_parser()

    args = parser.parse_args(sys.argv[1:])
    if hasattr(args, 'run'):
        args.run(args)
    else:
        parser.print_usage()


def test():
    import pytest
    path = os.path.dirname(__file__)
    test_path = os.path.join(path, 'tests')
    pytest.main(sys.argv[2:] + [test_path])


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == 'test':
        test()
    else:
        main()



#  tests for sputnik-0.9.3-py39hde42818_1002 (this is a generated file);
print('===== testing package: sputnik-0.9.3-py39hde42818_1002 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
import pytest
import sputnik

PACKAGE_DIR = os.path.abspath(os.path.dirname(sputnik.__file__))
pytest.main([PACKAGE_DIR])
#  --- run_test.py (end) ---

print('===== sputnik-0.9.3-py39hde42818_1002 OK =====');



import os

import pytest

from .. import install, build, remove, search, find, upload, update, files, purge


def test_build(sample_package_path):
    archive = build(sample_package_path)
    assert os.path.isfile(archive.path)


def test_build_install_remove(sample_package_path, tmp_path):
    archive = build(sample_package_path)
    assert os.path.isfile(archive.path)

    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 0

    package = install(None, None, archive.path, data_path=tmp_path)
    assert os.path.isdir(package.path)

    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 1
    assert packages[0].ident == package.ident

    remove(None, None, package.name, data_path=tmp_path)
    assert not os.path.isdir(package.path)

    packages = find(None, None, data_path=tmp_path)
    assert len(packages) == 0


def test_install_incompatible(sample_package_path, tmp_path):
    archive = build(sample_package_path)
    package = install(None, None, archive.path, data_path=tmp_path)

    packages = find(None, None, '%s ==%s, >=%s ,<=%s' % tuple([archive.name] + [archive.version]*3), data_path=tmp_path)
    assert len(packages) == 1
    assert packages[0].ident == archive.ident

    packages = find(None, None, '%s>%s,<%s' % tuple([archive.name] + [archive.version]*2), data_path=tmp_path)
    assert len(packages) == 0


def test_purge_raw(sample_package_path, tmp_path):
    assert len(find('test', '1.0.0', data_path=tmp_path)) == 0

    purge('test', '1.0.0', data_path=tmp_path)

    assert os.path.isdir(tmp_path)


def test_purge(sample_package_path, tmp_path):
    archive = build(sample_package_path)
    install('test', '1.0.0', archive.path, data_path=tmp_path)

    assert len(find('test', '1.0.0', data_path=tmp_path)) == 1

    purge('test', '1.0.0', data_path=tmp_path)

    assert len(find('test', '1.0.0', data_path=tmp_path)) == 0



import gzip
import io
import os
import tarfile
import time
import tempfile
import hashlib
import shutil
import logging

from . import util
from . import default


class EmptyArchiveException(Exception): pass
class InvalidPathException(Exception): pass


class ArchiveWriter(object):
    def __init__(self, path, base_path=None):
        self.logger = logging.getLogger(__name__)

        self.base_path = base_path
        self.path = path
        self.tmp_path = tempfile.mkdtemp()
        self.tmp_archive_path = os.path.join(self.tmp_path, default.ARCHIVE_FILENAME)
        self.archive = io.open(self.tmp_archive_path, 'wb')
        self.meta = {'manifest': []}

    def cleanup(self):
        shutil.rmtree(self.tmp_path)
        assert not os.path.exists(self.tmp_path)

    def close(self):
        self.archive.close()

        if not self.meta['manifest']:
            self.cleanup()
            raise EmptyArchiveException(self.path)

        mtime = time.time()

        tar = tarfile.open(self.path, 'w')
        with io.open(self.tmp_archive_path, 'rb') as f:
            checksum = hashlib.md5(f.read()).hexdigest()
            f.seek(os.SEEK_SET, 0)

            info = tarfile.TarInfo(default.ARCHIVE_FILENAME)
            info.size = os.stat(self.tmp_archive_path).st_size
            info.mtime = os.stat(self.tmp_archive_path).st_mtime
            info.type = tarfile.REGTYPE
            info.mode = 0o644
            info.uid = info.gid = 1000
            tar.addfile(info, f)
            self.meta['archive'] = (default.ARCHIVE_FILENAME, checksum)

        data = util.json_dump(self.meta)
        with io.BytesIO(data) as f:
            info = tarfile.TarInfo(default.META_FILENAME)
            info.size = len(data)
            info.mtime = mtime
            info.type = tarfile.REGTYPE
            info.mode = 0o644
            info.uid = info.gid = 1000
            tar.addfile(info, f)
        tar.close()

        self.cleanup()

    def add_json(self, name, obj):
        self.logger.info("adding %s", name)
        self.meta[name] = obj

    def add(self, path, cb=None):
        self.logger.info("adding %s", path)
        if self.base_path is None and os.path.isabs(path):
            raise InvalidPathException('cannot handle absolute paths without base_path: %s' % path)

        checksum = hashlib.md5()

        gz = gzip.GzipFile(fileobj=self.archive,
                           compresslevel=default.COMPRESSLEVEL)
        with io.open(path, 'rb') as f:
            bytes_read = 0

            while True:
                chunk = f.read(default.CHUNK_SIZE)
                if not chunk:
                    break

                bytes_read += len(chunk)

                gz.write(chunk)
                checksum.update(chunk)

                # callback for progress tracking
                if cb:
                    cb(bytes_read)

        gz.close()

        meta_path = os.path.relpath(path, self.base_path or '').split(os.path.sep)

        self.meta['manifest'].append({
            'path': meta_path,
            'noffset': self.archive.tell(),
            'size': os.stat(path).st_size,
            'checksum': (checksum.name, checksum.hexdigest()),
        })

    def add_path(self, path, cb=None):
        for root, _, filenames in os.walk(path):
            for filename in filenames:
                file_path = os.path.join(root, filename)
                self.add(file_path, cb=cb)
